/*
 *  Copyright (C) 2021 Texas Instruments Incorporated
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <kernel/dpl/DebugP.h>
#include <kernel/dpl/ClockP.h>
#include <kernel/dpl/AddrTranslateP.h>
#include <kernel/dpl/HwiP.h>
#include "ti_drivers_config.h"
#include "ti_drivers_open_close.h"
#include "ti_board_open_close.h"

#include <drivers/sciclient.h>

/*
 * This example configures a GPIO pin in input mode
 * and configures it to generate interrupt on rising edge.
 * The application waits for 5 key presses, prints the
 * number of times the keys are pressed and exits.
 */

#define R5F_VIM_INT_ID1                             ( 106 )

uint32_t            gGpioBaseAddr = GPIO_PUSH_BUTTON_BASE_ADDR;
HwiP_Object         gGpioHwiObject;
volatile uint32_t   gGpioIntrDone = 0;

uint32_t            gMcuGpio0BaseAddr = CONFIG_MCU_GPIO0_B0_IP0_BASE_ADDR;
HwiP_Object         gMcuGpio0Bank0HwiObject;

uint32_t gMcuGpio0Bank0Pin0Num;

uint32_t gMcuGpio0Bank0IsrCnt=0;
volatile uint32_t gMcuGpio0Bank0Pin0IsrCnt=0;

static void GPIO_bankIsrFxn(void *args);
static void MCUGPIO_bank0IsrFxn(void *args);

extern void Board_gpioInit(void);
extern void Board_gpioDeinit(void);

extern int32_t gpioIrInit(
    uint16_t src_id,
    uint16_t src_index,
    uint16_t dst_id,
    uint16_t dst_host_irq
);

extern int32_t gpioIrDeInit(
    uint16_t src_id,
    uint16_t src_index,
    uint16_t dst_id,
    uint16_t dst_host_irq
);

extern uint32_t Board_getGpioButtonIntrNum(void);
extern uint32_t Board_getGpioButtonSwitchNum(void);

void gpio_input_interrupt_main(void *args)
{
    int32_t         retVal;
    uint32_t        pinNum, intrNum;
    uint32_t        bankNum, waitCount = 5;
    HwiP_Params     hwiPrms;
    uint32_t        mcuGpio0Bank0Num;
    Bool            done;

    /* Open drivers to open the UART driver for console */
    Drivers_open();
    Board_driversOpen();
    Board_gpioInit();

    retVal = gpioIrInit(TISCI_DEV_MCU_GPIO0, 30, TISCI_DEV_R5FSS0_CORE0, R5F_VIM_INT_ID1);
    if (retVal != 0)
    {
        DebugP_assert(FALSE);
    }

    DebugP_log("GPIO Input Interrupt Test Started ...\r\n");
    DebugP_log("GPIO Interrupt Configured for Rising Edge (Button release will trigger interrupt) ...\r\n");

    pinNum          = GPIO_PUSH_BUTTON_PIN;
    intrNum         = Board_getGpioButtonIntrNum();
    bankNum         = GPIO_GET_BANK_INDEX(pinNum);

    /* Get MCU GPIO0 Bank0 pin number */
    gMcuGpio0Bank0Pin0Num  = CONFIG_MCU_GPIO0_B0_IP0_PIN;
    /* Get MCU GPIO Bank0 number */
    mcuGpio0Bank0Num   = GPIO_GET_BANK_INDEX(gMcuGpio0Bank0Pin0Num);

    /* Address translate */
    gGpioBaseAddr = (uint32_t) AddrTranslateP_getLocalAddr(gGpioBaseAddr);
    gMcuGpio0BaseAddr = (uint32_t) AddrTranslateP_getLocalAddr(gMcuGpio0BaseAddr);

    /* Setup GPIO for interrupt generation */
    GPIO_setDirMode(gGpioBaseAddr, pinNum, GPIO_PUSH_BUTTON_DIR);
    GPIO_setTrigType(gGpioBaseAddr, pinNum, GPIO_PUSH_BUTTON_TRIG_TYPE);

    /* Setup MCU GPIO0 Bank0 GPIOs for interrupt generation */
    GPIO_setDirMode(gMcuGpio0BaseAddr, gMcuGpio0Bank0Pin0Num, CONFIG_MCU_GPIO0_B0_IP0_DIR);
    GPIO_setTrigType(gMcuGpio0BaseAddr, gMcuGpio0Bank0Pin0Num, CONFIG_MCU_GPIO0_B0_IP0_TRIG_TYPE);

    /* Register pin interrupt */
    HwiP_Params_init(&hwiPrms);
    hwiPrms.intNum   = intrNum;
    hwiPrms.callback = &GPIO_bankIsrFxn;
    hwiPrms.args     = (void *) pinNum;
    retVal = HwiP_construct(&gGpioHwiObject, &hwiPrms);
    DebugP_assert(retVal == SystemP_SUCCESS );

    /* Register pin interrupt */
    HwiP_Params_init(&hwiPrms);
    hwiPrms.intNum   = R5F_VIM_INT_ID1;
    hwiPrms.callback = &MCUGPIO_bank0IsrFxn;
    hwiPrms.args     = (void *) mcuGpio0Bank0Num;
    retVal = HwiP_construct(&gMcuGpio0Bank0HwiObject, &hwiPrms);
    DebugP_assert(retVal == SystemP_SUCCESS );

    /* Enable GPIO Bank interrupts */
    GPIO_bankIntrEnable(gGpioBaseAddr, bankNum);
    GPIO_bankIntrEnable(gMcuGpio0BaseAddr, mcuGpio0Bank0Num);

    done = FALSE;
    while (!done)
    {
        /* Keep printing the current GPIO value */
        DebugP_log("Key is pressed %d times\r\n", gGpioIntrDone);
        DebugP_log("MCU_GPIO0_%d triggered %d times\r\n", gMcuGpio0Bank0Pin0Num, gMcuGpio0Bank0Pin0IsrCnt);

        ClockP_sleep(1);

        done = ((gGpioIntrDone >= waitCount) &&
                (gMcuGpio0Bank0Pin0IsrCnt >= waitCount));
    }
    DebugP_log("Key is pressed %d times\r\n", gGpioIntrDone);
    DebugP_log("MCU_GPIO0_%d triggered %d times\r\n", gMcuGpio0Bank0Pin0Num, gMcuGpio0Bank0Pin0IsrCnt);

    /* Unregister interrupt */
    GPIO_bankIntrDisable(gGpioBaseAddr, bankNum);
    GPIO_setTrigType(gGpioBaseAddr, pinNum, GPIO_TRIG_TYPE_NONE);
    GPIO_clearIntrStatus(gGpioBaseAddr, pinNum);
    HwiP_destruct(&gGpioHwiObject);

    GPIO_bankIntrDisable(gMcuGpio0BaseAddr, mcuGpio0Bank0Num);
    GPIO_setTrigType(gMcuGpio0BaseAddr, gMcuGpio0Bank0Pin0Num, GPIO_TRIG_TYPE_NONE);
    GPIO_clearIntrStatus(gMcuGpio0BaseAddr, gMcuGpio0Bank0Pin0Num);
    HwiP_destruct(&gMcuGpio0Bank0HwiObject);

    DebugP_log("GPIO Input Interrupt Test Passed!!\r\n");
    DebugP_log("All tests have passed!!\r\n");

    Board_gpioDeinit();
    retVal = gpioIrDeInit(TISCI_DEV_MCU_GPIO0, 30, TISCI_DEV_R5FSS0_CORE0, 106);
    if (retVal != 0)
    {
        DebugP_assert(FALSE);
    }

    Board_driversClose();
    Drivers_close();
}

static void GPIO_bankIsrFxn(void *args)
{
    uint32_t    pinNum = (uint32_t) args;
    uint32_t    bankNum =  GPIO_GET_BANK_INDEX(pinNum);
    uint32_t    intrStatus, pinMask = GPIO_GET_BANK_BIT_MASK(pinNum);

    /* Get and clear bank interrupt status */
    intrStatus = GPIO_getBankIntrStatus(gGpioBaseAddr, bankNum);
    GPIO_clearBankIntrStatus(gGpioBaseAddr, bankNum, intrStatus);

    /* Per pin interrupt handling */
    if(intrStatus & pinMask)
    {
        gGpioIntrDone++;
    }
}

static void MCUGPIO_bank0IsrFxn(void *args)
{
    uint32_t    bankNum =  (uint32_t) args;
    uint32_t    intrStatus, pinMask;

    gMcuGpio0Bank0IsrCnt++;

    /* Get and clear bank interrupt status */
    intrStatus = GPIO_getBankIntrStatus(gMcuGpio0BaseAddr, bankNum);

    /* Per pin interrupt handling */
    while (intrStatus != 0)
    {
        GPIO_clearBankIntrStatus(gMcuGpio0BaseAddr, bankNum, intrStatus);

        /* Pin 0 */
        pinMask = GPIO_GET_BANK_BIT_MASK(gMcuGpio0Bank0Pin0Num);
        if (intrStatus & pinMask)
        {
            intrStatus &= ~pinMask;
            gMcuGpio0Bank0Pin0IsrCnt++;
        }

        intrStatus = GPIO_getBankIntrStatus(gMcuGpio0BaseAddr, bankNum);
    }
}

